//
// C++ Implementation: bot_util
//
// Description: Basic util functions, mostly bot related
//
//
// Author:  <rickhelmus@gmail.com>, (C) 2004
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "cube.h"

// Function code by PMB - Begin


// Random functions begin
// copied from
// http://forums.bots-united.com/showthread.php?t=244&page=1&pp=10&highlight=random
// So credits to PMB and the person(s) who actually made it :]

// maximum value returned by our number generator (2^31 - 1 = 0x7FFFFFFF)
#define LRAND_MAX 2147483647L
long lseed; // our random number generator's seed

long lrand (void)
{
	// this function is the equivalent of the rand() standard C library function, except that
	// whereas rand() works only with short integers (i.e. not above 32767), this function is
	// able to generate 32-bit random numbers. Isn't that nice ?
	// credits go to Ray Gardner for his fast implementation of minimal random number generators
	// http://c.snippets.org/snip_lister.php?fname=rg_rand.c
	
	// compose the two 16-bit parts of the long integer and assemble them
	static unsigned long lrand_lo, lrand_hi;
	lrand_lo = 16807 * (long) (lseed & 0xFFFF); // low part
	lrand_hi = 16807 * (long) ((unsigned long) lseed >> 16);
	// high part
	lrand_lo += (lrand_hi & 0x7FFF) << 16;
	// assemble both in lrand_lo
	// is the resulting number greater than LRAND_MAX (half the capacity) ?
	if (lrand_lo > LRAND_MAX)
	{
		lrand_lo &= LRAND_MAX; // then get rid of the disturbing bit
		lrand_lo++; // and increase it a bit (to avoid overflow problems, I suppose)
	}
	
	lrand_lo += lrand_hi >> 15;
	// now do twisted maths to generate the next seed
	// is the resulting number greater than LRAND_MAX (half the capacity) ?
	if (lrand_lo > LRAND_MAX)
	{
		lrand_lo &= LRAND_MAX; // then get rid of the disturbing bit
		lrand_lo++; // and increase it a bit (to avoid overflow problems, I suppose)
	}
	// now we've got our (pseudo-)random number.
	lseed = (long) lrand_lo; // put it in the seed for next time
	return (lseed); // and return it. Yeah, simple as that.
}

void lsrand (unsigned long initial_seed)
{
	// this function initializes the random seed based on the initial seed value passed in the
	// initial_seed parameter. Since random seeds are usually initialized with the time of day,
	// and time is a value that changes slowly, we bump the seed twice to have it in a more
	// random state for future calls of the random number generator.
	lseed = (long) initial_seed; // fill in the initial seed of the random number generator
	lseed = lrand (); // bump it once
	lseed = lrand (); // bump it twice
	return; // that's all folks
}

long RandomLong (long from, long to)
{
	// this function returns a random integer number between (and including) the starting and
	// ending values passed by parameters from and to.
	if (to <= from) 
		return (from);
		
	return (from + lrand () / (LRAND_MAX / (to - from + 1)));
}

float RandomFloat (float from, float to)
{
	// this function returns a random floating-point number between (and including) the starting
	// and ending values passed by parameters from and to.
	
	if (to <= from)
		return (from);
		
	return (from + (float) lrand () / (LRAND_MAX / (to - from)));
}

// End random functions

void AnglesToVectors(vec angles, vec &forward, vec &right, vec &up)
{
     static float degrees_to_radians = 2 * PI / 360;
     float angle;
     float sin_pitch;
     float sin_yaw;
     float sin_roll;
     float cos_pitch;
     float cos_yaw;
     float cos_roll;
         
     // For some reason this has to be done :)
     // Me = math n00b
     angles.x = -angles.x;
     angles.y -= 90;
     
     // compute the sin and cosine of the pitch component
     angle = angles.x * degrees_to_radians;
     sin_pitch = sinf (angle);
     cos_pitch = cosf (angle);
     
     // compute the sin and cosine of the yaw component
     angle = angles.y * degrees_to_radians;
     sin_yaw = sinf (angle);
     cos_yaw = cosf (angle);
     
     // compute the sin and cosine of the roll component
     angle = angles.z * degrees_to_radians;
     sin_roll = sinf (angle);
     cos_roll = cosf (angle);
     
     // build the FORWARD vector
     forward.x = cos_pitch * cos_yaw;
     forward.y = cos_pitch * sin_yaw;
     forward.z = -sin_pitch;
     
     // build the RIGHT vector
     right.x = (-(sin_roll * sin_pitch * cos_yaw) - (cos_roll * -sin_yaw));
     right.y = (-(sin_roll * sin_pitch * sin_yaw) - (cos_roll * cos_yaw));
     right.z = -(sin_roll * cos_pitch);
           
     // build the UPWARDS vector
     up.x = ((cos_roll * sin_pitch * cos_yaw) - (sin_roll * -sin_yaw));
     up.y = ((cos_roll * sin_pitch * sin_yaw) - (sin_roll * cos_yaw));
     up.z = cos_roll * cos_pitch;
     
     return;
}

// Function code by PMB - End

vec WrapAngles (vec angles) 
{
   // check for wraparound of angles
   if (angles.x > 90) angles.x = 90;
   else if (angles.x < 90) angles.x = -90;
   while (angles.y > 360) angles.y -= 360;
   while (angles.y < 0) angles.y += 360;
   while (angles.z > 360) angles.z -= 360;
   while (angles.z < 0) angles.z += 360;

   return (angles);
}

float WrapXAngle(float angle)
{
     if (angle > 90) angle = 90;
     else if (angle < 90) angle = -90;

     return angle;
}

float WrapYZAngle(float angle)
{
     while (angle >= 360.0f) angle -= 360.0f;
     while (angle < 0.0f) angle += 360.0f;

     return angle;
}

// Sees how far it can come from 'from' to 'to' and puts the end vector in 'end'
// (1337 description :-)
// UNDONE: Optimize this as much as possible
void TraceLine(vec from, vec to, vec &end, dynent *pTracer, bool CheckPlayers)
{
    end=from;
    
    if(OUTBORD((int)from.x, (int)from.y) || OUTBORD((int)to.x, (int)to.y)) return;
    float dx = to.x-from.x;
    float dy = to.y-from.y; 
    int steps = (int)(sqrt(dx*dx+dy*dy)/0.9);
    if(!steps) return;
    float x = from.x;
    float y = from.y;
    int i = 0;
    const float tracerradius = (pTracer) ? pTracer->radius : 0;
    bool bCollided;
    
    for(;;)
    {
         // First check if the 'line' collides with any cubes
         sqr *s = S(fast_f2nat(x), fast_f2nat(y));
         if(SOLID(s)) break;
         float floor = s->floor;
         if(s->type==FHF) floor -= s->vdelta/4.0f;
         float ceil = s->ceil;
         if(s->type==CHF) ceil += s->vdelta/4.0f;
         float rz = from.z-((from.z-to.z)*(i/(float)steps));
         if(rz<floor || rz>ceil) break;
        
         bCollided = false;
         
         // Now check if the 'line' collides with entities like mapmodels
         loopv(ents)
         {
              entity &e = ents[i];
              if(e.type!=MAPMODEL) continue; // Only map models for now
              mapmodelinfo &mmi = getmminfo(e.attr2);
              if(!&mmi || !mmi.h) continue;
              const float r = mmi.rad+tracerradius;
              if(fabs(e.x-x)<r && fabs(e.y-y)<r)
              {
                   vec eorig = {e.x, e.y, e.z};
                   vec curpos = {x, y, rz};
                   vdist(flDist, vecto, curpos, eorig);
                   float pitch = atan2(e.z-rz, flDist)*180/PI;
                   float yaw = -(float)atan2(e.x - x, e.y - y)/PI*180+180;
                   vec angles = { pitch, yaw, 0 };
                   vec forward, right, up;
             
                   AnglesToVectors(WrapAngles(angles), forward, right, up);
             
                   end = eorig;
                   vmul(forward, mmi.rad);
                   vadd(end, forward);
                   
                   if (!bCollided)
		   {
		   	bCollided = true;
			break;
		   }
              }
         }

	 if (!bCollided && CheckPlayers)
	 {         
         	// Now check if the 'line' collides with any players
         	loopv(players)
         	{
              		dynent *d = players[i];
                        if (!d || (d==pTracer)) continue;
              		const float r = d->radius+tracerradius;
              		if(fabs(d->o.x-x)<r && fabs(d->o.y-y)<r)
              		{
                   		vec curpos = {x, y, rz};
                   		vdist(flDist, vecto, curpos, d->o);
                   		float pitch = atan2(d->o.z-rz, flDist)*180/PI;
                   		float yaw = -(float)atan2(d->o.x - x, d->o.y - y)/PI*180+180;
                   		vec angles = { pitch, yaw, 0 };
                   		vec forward, right, up;
             
                   		AnglesToVectors(WrapAngles(angles), forward, right, up);
             
                   		end = d->o;
                   		vmul(forward, d->radius);
                   		vadd(end, forward);
                   
                   		if (!bCollided)
		   		{
		   			bCollided = true;
					break;
		   		}
              		}
         	}
         }
	 
	 if (!bCollided && CheckPlayers)
	 {         
         	// Now check if the 'line' collides with any bots
         	loopv(bots)
         	{
              		dynent *d = bots[i];
                        if (!d || (d==pTracer)) continue;
              		const float r = d->radius+tracerradius;
              		if(fabs(d->o.x-x)<r && fabs(d->o.y-y)<r)
              		{
                   		vec curpos = {x, y, rz};
                   		vdist(flDist, vecto, curpos, d->o);
                   		float pitch = atan2(d->o.z-rz, flDist)*180/PI;
                   		float yaw = -(float)atan2(d->o.x - x, d->o.y - y)/PI*180+180;
                   		vec angles = { pitch, yaw, 0 };
                   		vec forward, right, up;
             
                   		AnglesToVectors(WrapAngles(angles), forward, right, up);
             
                   		end = d->o;
                   		vmul(forward, d->radius);
                   		vadd(end, forward);
                   
                   		if (!bCollided)
		   		{
		   			bCollided = true;
					break;
		   		}
              		}
         	}
         }
	 
	 if (bCollided)
              break;
              
         end.x = x;
         end.y = y;
         end.z = rz;
         x += dx/(float)steps;
         y += dy/(float)steps;
         i++;
    }
       
    return;
}
